@@ -1,5 +1,5 @@
 /* msgfmt - converts Uniforum style .po files to binary .mo files
-   Copyright (C) 1995 Free Software Foundation, Inc.
+   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
    Written by Ulrich Drepper <drepper@gnu.ai.mit.edu>, April 1995.
 
 This program is free software; you can redistribute it and/or modify
@@ -37,6 +37,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
 #include "hash.h"
 
+#include "dir-list.h"
 #include "error.h"
 #include "getline.h"
 #include "printf.h"
@@ -46,6 +47,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 #include "domain.h"
 #include "hash-string.h"
 #include <libintl.h>
+#include "message.h"
 #include "po.h"
 
 #define _(str) gettext (str)
@@ -69,12 +71,30 @@ struct msgstr_def
   lex_pos_ty pos;
 };
 
+/* This structure defines a derived class of the po_ty class.  (See
+   po.h for an explanation.)  */
+typedef struct msgfmt_class_ty msgfmt_class_ty;
+struct msgfmt_class_ty
+{
+  /* inherited instance variables, etc */
+  PO_BASE_TY
+
+  int is_fuzzy;
+  enum is_c_format is_c_format;
+  enum is_c_format do_wrap;
+
+  int has_header_entry;
+};
+
 /* Alignment of strings in resulting .mo file.  */
 static size_t alignment;
 
 /* Contains exit status for case in which no premature exit occurs.  */
 static int exit_status;
 
+/* If nonzero include even fuzzy translations in output file.  */
+static int include_all;
+
 /* Nonzero if no hash table in .mo is wanted.  */
 static int no_hash_table;
 
@@ -91,17 +111,33 @@ static struct msg_domain *domain;
 static struct msg_domain *current_domain;
 
 /* If not zero list duplicate message identifiers.  */
-static int verbose;
+static int verbose_level;
+
+/* If not zero check strings according to format string rules for the
+   language.  */
+static int do_check;
+
+/* Counters for statistics on translations for the processed files.  */
+static int msgs_translated;
+static int msgs_untranslated;
+static int msgs_fuzzy;
+
+/* If not zero print statistics about translation at the end.  */
+static int do_statistics;
 
 /* Long options.  */
 static const struct option long_options[] =
 {
   { "alignment", required_argument, NULL, 'a' },
+  { "check", no_argument, &do_check, 1 },
+  { "directory", required_argument, NULL, 'D' },
   { "help", no_argument, NULL, 'h' },
   { "no-hash", no_argument, &no_hash_table, 1 },
   { "output-file", required_argument, NULL, 'o' },
+  { "statistics", no_argument, &do_statistics, 1 },
   { "strict", no_argument, NULL, 'S' },
-  { "verbose", no_argument, &verbose, 1 },
+  { "use-fuzzy", no_argument, NULL, 'f' },
+  { "verbose", no_argument, NULL, 'v' },
   { "version", no_argument, NULL, 'V' },
   { NULL, 0, NULL, 0 }
 };
@@ -117,25 +153,27 @@ static const struct option long_options[] =
 #endif /* roundup  */
 
 /* Prototypes for local functions.  */
-static void usage __P ((int status))
+static void usage PARAMS ((int status))
 #if defined __GNUC__ && ((__GNUC__ == 2 && __GNUC_MINOR__ >= 5) || __GNUC__ > 2)
 	__attribute__ ((noreturn))
 #endif
 ;
-static void error_print __P ((void));
-static void grammar __P ((char *__filename));
-static void format_directive_domain __P ((po_ty *__pop, char *__name));
-static void format_directive_message __P ((po_ty *__pop, char *__msgid,
-					   lex_pos_ty *__msgid_pos,
-					   char *__msgstr,
-					   lex_pos_ty *__msgstr_pos));
-static struct msg_domain *new_domain __P ((const char *name));
-static int compare_id __P ((const void *pval1, const void *pval2));
-static void write_table __P ((FILE *output_file, hash_table *tab));
-static void check_pair __P ((const char *msgid, const lex_pos_ty *msgid_pos,
-			     const char *msgstr,
-			     const lex_pos_ty *msgstr_pos));
-static const char *add_mo_suffix __P ((const char *));
+static void error_print PARAMS ((void));
+static void grammar PARAMS ((char *__filename));
+static void format_constructor PARAMS ((po_ty *__that));
+static void format_directive_domain PARAMS ((po_ty *__pop, char *__name));
+static void format_directive_message PARAMS ((po_ty *__pop, char *__msgid,
+					      lex_pos_ty *__msgid_pos,
+					      char *__msgstr,
+					      lex_pos_ty *__msgstr_pos));
+static void format_comment_special PARAMS ((po_ty *pop, const char *s));
+static struct msg_domain *new_domain PARAMS ((const char *name));
+static int compare_id PARAMS ((const void *pval1, const void *pval2));
+static void write_table PARAMS ((FILE *output_file, hash_table *tab));
+static void check_pair PARAMS ((const char *msgid, const lex_pos_ty *msgid_pos,
+				const char *msgstr,
+				const lex_pos_ty *msgstr_pos, int is_format));
+static const char *add_mo_suffix PARAMS ((const char *));
 
 
 int
@@ -143,7 +181,7 @@ main(argc, argv)
      int argc;
      char *argv[];
 {
-  int optchar;
+  int opt;
   int do_help = 0;
   int do_version = 0;
   int strict_uniforum = 0;
@@ -154,6 +192,7 @@ main(argc, argv)
   /* Set program name for messages.  */
   program_name = argv[0];
   error_print_progname = error_print;
+  error_one_per_line = 1;
   exit_status = EXIT_SUCCESS;
 
 #ifdef HAVE_SETLOCALE
@@ -165,11 +204,11 @@ main(argc, argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
-  while ((optchar = getopt_long (argc, argv, "a:ho:vV", long_options, NULL))
+  while ((opt = getopt_long (argc, argv, "a:cD:fho:vV", long_options, NULL))
 	 != EOF)
-    switch (optchar)
+    switch (opt)
       {
-      case '\0':
+      case '\0':		/* Long option.  */
 	break;
       case 'a':
 	{
@@ -180,6 +219,15 @@ main(argc, argv)
 	    alignment = new_align;
 	}
 	break;
+      case 'c':
+	do_check = 1;
+	break;
+      case 'D':
+	dir_list_append (optarg);
+	break;
+      case 'f':
+	include_all = 1;
+	break;
       case 'h':
 	do_help = 1;
 	break;
@@ -190,7 +238,7 @@ main(argc, argv)
 	strict_uniforum = 1;
 	break;
       case 'v':
-	verbose = 1;
+	++verbose_level;
 	break;
       case 'V':
 	do_version = 1;
@@ -223,6 +271,11 @@ main(argc, argv)
   if (output_file_name != NULL)
     current_domain = new_domain (output_file_name);
 
+  /* Prepare PO file reader.  We need to see the comments because inexact
+     translations must be reported.  */
+  po_lex_pass_comments (1);
+
+  /* Now write out all domains.  */
   /* Process all given .po files.  */
   while (argc > optind)
     {
@@ -237,7 +290,6 @@ main(argc, argv)
       ++optind;
     }
 
-  /* Now write out all domains.  */
   while (domain != NULL)
     {
       FILE *output_file;
@@ -277,6 +329,17 @@ main(argc, argv)
       domain = domain->next;
     }
 
+  /* Print statistics if requested.  */
+  if (verbose_level > 0 || do_statistics)
+    {
+      fprintf (stderr, _("%d translated messages"), msgs_translated);
+      if (msgs_fuzzy > 0)
+	fprintf (stderr, _(", %d fuzzy translations"), msgs_fuzzy);
+      if (msgs_untranslated > 0)
+	fprintf (stderr, _(", %d untranslated messages"), msgs_untranslated);
+      fputs (".\n", stderr);
+    }
+
   exit (exit_status);
 }
 
@@ -290,16 +353,23 @@ usage (status)
     fprintf (stderr, _("Try `%s --help' for more information\n"),
 	     program_name);
   else
+    /* xgettext: no-wrap */
     printf (_("\
 Usage: %s [OPTION] filename.po ...\n\
 Mandatory arguments to long options are mandatory for short options too.\n\
-  -a, --alignment=NUMBER   align strings to NUMBER bytes (default: %d)\n\
-  -h, --help               display this help and exit\n\
-      --no-hash            binary file will not include the hash table\n\
-  -o, --output-file=FILE   specify output file name as FILE\n\
-      --strict             enable strict Uniforum mode\n\
-  -v, --verbose            list input file anomalies\n\
-  -V, --version            output version information and exit\n\
+  -a, --alignment=NUMBER      align strings to NUMBER bytes (default: %d)\n\
+  -c, --check                 perform language dependent checks on strings\n\
+  -D, --directory=DIRECTORY   add DIRECTORY to list for input files search\n\
+  -f, --use-fuzzy             use fuzzy entries in output\n\
+  -h, --help                  display this help and exit\n\
+      --no-hash               binary file will not include the hash table\n\
+  -o, --output-file=FILE      specify output file name as FILE\n\
+      --statistics            print statistics about translations\n\
+      --strict                enable strict Uniforum mode\n\
+  -v, --verbose               list input file anomalies\n\
+  -V, --version               output version information and exit\n\
+\n\
+Giving the -v option more than once increases the verbosity level.\n\
 \n\
 If input file is -, standard input is read.  If output file is -,\n\
 output is written to standard output.\n"),
@@ -342,6 +412,33 @@ error_print ()
 }
 
 
+/* Prepare for first message.  */
+static void
+format_constructor (that)
+     po_ty *that;
+{
+  msgfmt_class_ty *this = (msgfmt_class_ty *) that;
+
+  this->is_fuzzy = 0;
+  this->is_c_format = undecided;
+  this->do_wrap = undecided;
+  this->has_header_entry = 0;
+}
+
+
+/* Some checks after whole file is read.  */
+static void
+format_debrief (that)
+     po_ty *that;
+{
+  msgfmt_class_ty *this = (msgfmt_class_ty *) that;
+
+  /* If in verbose mode, test whether header entry was found.  */
+  if (verbose_level > 0 && this->has_header_entry == 0)
+    error (0, 0, _("warning: no header entry found"));
+}
+
+
 /* Process `domain' directive from .po file.  */
 static void
 format_directive_domain (pop, name)
@@ -375,7 +472,7 @@ domain name \"%s\" not suitable as file name: will use prefix"), name);
     }
   else
     {
-      if (verbose)
+      if (verbose_level > 0)
 	/* We don't change the exit status here because this is really
 	   only an information.  */
 	error (0, 0, _("`domain %s' directive ignored"), name);
@@ -388,31 +485,99 @@ domain name \"%s\" not suitable as file name: will use prefix"), name);
 
 /* Process `msgid'/`msgstr' pair from .po file.  */
 static void
-format_directive_message (pop, msgid_string, msgid_pos, msgstr_string,
+format_directive_message (that, msgid_string, msgid_pos, msgstr_string,
 			  msgstr_pos)
-     po_ty *pop;
+     po_ty *that;
      char *msgid_string;
      lex_pos_ty *msgid_pos;
      char *msgstr_string;
      lex_pos_ty *msgstr_pos;
 {
+  msgfmt_class_ty *this = (msgfmt_class_ty *) that;
   struct msgstr_def *entry;
 
-  if (msgstr_string[0] == '\0')
+  if (msgstr_string[0] == '\0' || (!include_all && this->is_fuzzy))
     {
-      if (verbose)
+      if (verbose_level > 1)
 	/* We don't change the exit status here because this is really
 	   only an information.  */
-	error_with_loc (0, 0, msgstr_pos->file_name, msgstr_pos->line_number,
-			_("empty `msgstr' entry ignored"));
+	error_at_line (0, 0, msgstr_pos->file_name, msgstr_pos->line_number,
+		       (msgstr_string[0] == '\0'
+			? _("empty `msgstr' entry ignored")
+			: _("fuzzy `msgstr' entry ignored")));
 
       /* Free strings allocated in po-gram.y.  */
-      free (msgid_string);
       free (msgstr_string);
 
-      return;
+      /* Increment counter for untranslated messages.  */
+      ++msgs_untranslated;
+
+      goto prepare_next;
     }
 
+  /* Test for header entry.  */
+  if (msgid_string[0] == '\0')
+    {
+      this->has_header_entry = 1;
+
+      /* Do some more tests on test contents of the header entry.  */
+      if (verbose_level > 0)
+	{
+	  static const char *required_fields[] =
+	  {
+	    "Project-Id-Version", "PO-Revision-Date",
+	    "Last-Translator", "Language-Team", "MIME-Version",
+	    "Content-Type", "Content-Transfer-Encoding"
+	  };
+	  static const char *default_values[] =
+	  {
+	    "PACKAGE VERSION", "YEAR-MO-DA", "FULL NAME", "LANGUAGE",
+	    NULL, "text/plain; charset=CHARSET", "ENCODING"
+	  };
+	  const size_t nfields = (sizeof (required_fields)
+				  / sizeof (required_fields[0]));
+	  int initial = -1;
+	  int cnt;
+
+	  for (cnt = 0; cnt < nfields; ++cnt)
+	    {
+	      char *endp = strstr (msgstr_string, required_fields[cnt]);
+
+	      if (endp == NULL)
+		error (0, 0, _("headerfield `%s' missing in header"),
+		       required_fields[cnt]);
+	      else if (endp != msgstr_string && endp[-1] != '\n')
+		error (0, 0, _("\
+header field `%s' should start at beginning of line"),
+		       required_fields[cnt]);
+	      else if (default_values[cnt] != NULL
+		       && strncmp (default_values[cnt],
+				   endp + strlen (required_fields[cnt]) + 2,
+				   strlen (default_values[cnt])) == 0)
+		if (initial != -1)
+		  {
+		    error (0, 0, _("\
+some header fields still have the initial default value"));
+		    initial = -1;
+		    break;
+		  }
+		else
+		  initial = cnt;
+	    }
+
+	  if (initial != -1)
+	    error (0, 0, _("field `%s' still has initial default value"),
+		   required_fields[initial]);
+	}
+    }
+  else
+    /* We don't count the header entry in the statistic so place the
+       counter incrementation here.  */
+    if (this->is_fuzzy)
+      ++msgs_fuzzy;
+    else
+      ++msgs_translated;
+
   /* We found a valid pair of msgid/msgstr.
      Construct struct to describe msgstr definition.  */
   entry = (struct msgstr_def *) xmalloc (sizeof (*entry));
@@ -420,14 +585,14 @@ format_directive_message (pop, msgid_string, msgid_pos, msgstr_string,
   entry->msgstr = msgstr_string;
   entry->pos = *msgstr_pos;
 
-  /* If --verbose is selected we do some more checks on both strings.  */
-  if (verbose)
-    check_pair (msgid_string, msgid_pos, msgstr_string, msgstr_pos);
+  /* Do some more checks on both strings.  */
+  check_pair (msgid_string, msgid_pos, msgstr_string, msgstr_pos,
+	      do_check && possible_c_format_p (this->is_c_format));
 
   /* Check whether already a domain is specified.  If not use default
      domain.  */
   if (current_domain == NULL)
-    current_domain = new_domain ("messages"); /* FIXME */
+    current_domain = new_domain ("messages");
 
   /* We insert the ID/string pair into the hashing table.  But we have
      to take care for dublicates.  */
@@ -439,15 +604,15 @@ format_directive_message (pop, msgid_string, msgid_pos, msgstr_string,
       /* And the parameter string (allocated in po-gram.y).  */
       free (msgstr_string);
 
-      if (verbose)
+      if (verbose_level > 0)
 	{
 	  /* We give a warning about this.  Get the old definition for
 	     reference.  */
 	  find_entry (&current_domain->symbol_tab, msgid_string,
 		      strlen (msgid_string), (void **) &entry);
 
-	  gram_error_with_loc (msgid_pos, _("duplicate message ID"));
-	  gram_error_with_loc (&entry->pos, _("\
+	  gram_error_at_line (msgid_pos, _("duplicate message ID"));
+	  gram_error_at_line (&entry->pos, _("\
 this is the location of the first definition"));
 
 	  /* FIXME Should this be always a reason for an exit status != 0?  */
@@ -455,8 +620,40 @@ this is the location of the first definition"));
 	}
     }
 
+prepare_next:
   /* We do not need the msgid string in any case.  */
   free (msgid_string);
+
+  /* Prepare for next message.  */
+  this->is_fuzzy = 0;
+  this->is_c_format = undecided;
+  this->do_wrap = undecided;
+}
+
+
+/* Test for `#, fuzzy' comments and warn.  */
+static void
+format_comment_special (that, s)
+     po_ty *that;
+     const char *s;
+{
+  msgfmt_class_ty *this = (msgfmt_class_ty *) that;
+
+  if (strstr (s, "fuzzy") != NULL)
+    {
+      static int warned = 0;
+
+      if (!include_all && verbose_level > 1 && warned == 0)
+	{
+	  warned = 1;
+	  error (0, 0, _("\
+WARNING: source file contains fuzzy translation"));
+	}
+
+      this->is_fuzzy = 1;
+    }
+  this->is_c_format = parse_c_format_description_string (s);
+  this->do_wrap = parse_c_width_description_string (s);
 }
 
 
@@ -631,11 +828,12 @@ write_table (output_file, tab)
 
 
 static void
-check_pair (msgid, msgid_pos, msgstr, msgstr_pos)
+check_pair (msgid, msgid_pos, msgstr, msgstr_pos, is_format)
      const char *msgid;
      const lex_pos_ty *msgid_pos;
      const char *msgstr;
      const lex_pos_ty *msgstr_pos;
+     int is_format;
 {
   size_t msgid_len = strlen (msgid);
   size_t msgstr_len = strlen (msgstr);
@@ -649,46 +847,50 @@ check_pair (msgid, msgid_pos, msgstr, msgstr_pos)
   /* Test 1: check whether both or none of the strings begin with a '\n'.  */
   if (((msgid[0] == '\n') ^ (msgstr[0] == '\n')) != 0)
     {
-      error_with_loc (0, 0, msgid_pos->file_name, msgid_pos->line_number, _("\
-`msgid' and `msgstr' entry does not both begin with '\\n'"));
+      error_at_line (0, 0, msgid_pos->file_name, msgid_pos->line_number, _("\
+`msgid' and `msgstr' entries does not both begin with '\\n'"));
       exit_status = EXIT_FAILURE;
     }
 
   /* Test 2: check whether both or none of the strings end with a '\n'.  */
   if (((msgid[msgid_len - 1] == '\n') ^ (msgstr[msgstr_len - 1] == '\n')) != 0)
     {
-      error_with_loc (0, 0, msgid_pos->file_name, msgid_pos->line_number, _("\
-`msgid' and `msgstr' entry does not both end with '\\n'"));
+      error_at_line (0, 0, msgid_pos->file_name, msgid_pos->line_number, _("\
+`msgid' and `msgstr' entries does not both end with '\\n'"));
       exit_status = EXIT_FAILURE;
     }
 
-  /* Test 3: check whether both formats strings contain the same number
-     of format specifications.  */
-  nidfmts = parse_printf_format (msgid, 0, NULL);
-  nstrfmts = parse_printf_format (msgstr, 0, NULL);
-  if (nidfmts != nstrfmts)
+  if (is_format != 0)
     {
-      error_with_loc (0, 0, msgid_pos->file_name, msgid_pos->line_number, _("\
+      /* Test 3: check whether both formats strings contain the same
+	 number of format specifications.  */
+      nidfmts = parse_printf_format (msgid, 0, NULL);
+      nstrfmts = parse_printf_format (msgstr, 0, NULL);
+      if (nidfmts != nstrfmts)
+	{
+	  error_at_line (0, 0, msgid_pos->file_name, msgid_pos->line_number,
+			 _("\
 number of format specifications in `msgid' and `msgstr' does not match"));
-      exit_status = EXIT_FAILURE;
-    }
-  else
-    {
-      int *id_args = (int *) alloca (nidfmts * sizeof (int));
-      int *str_args = (int *) alloca (nstrfmts * sizeof (int));
-      size_t cnt;
+	  exit_status = EXIT_FAILURE;
+	}
+      else
+	{
+	  int *id_args = (int *) alloca (nidfmts * sizeof (int));
+	  int *str_args = (int *) alloca (nstrfmts * sizeof (int));
+	  size_t cnt;
 
-      (void) parse_printf_format (msgid, nidfmts, id_args);
-      (void) parse_printf_format (msgstr, nstrfmts, str_args);
+	  (void) parse_printf_format (msgid, nidfmts, id_args);
+	  (void) parse_printf_format (msgstr, nstrfmts, str_args);
 
-      for (cnt = 0; cnt < nidfmts; ++cnt)
-	if (id_args[cnt] != str_args[cnt])
-	  {
-	    error_with_loc (0, 0, msgid_pos->file_name, msgid_pos->line_number,
-			    _("\
+	  for (cnt = 0; cnt < nidfmts; ++cnt)
+	    if (id_args[cnt] != str_args[cnt])
+	      {
+		error_at_line (0, 0, msgid_pos->file_name,
+			       msgid_pos->line_number, _("\
 format specifications for argument %u are not the same"), cnt);
-	    exit_status = EXIT_FAILURE;
-	  }
+		exit_status = EXIT_FAILURE;
+	      }
+	}
     }
 }
 
@@ -701,16 +903,17 @@ format specifications for argument %u are not the same"), cnt);
 
 static po_method_ty format_methods =
 {
-  sizeof(po_ty),
-  NULL, /* constructor */
+  sizeof (msgfmt_class_ty),
+  format_constructor, /* constructor */
   NULL, /* destructor */
   format_directive_domain,
   format_directive_message,
   NULL, /* parse_brief */
-  NULL, /* parse_debrief */
+  format_debrief, /* parse_debrief */
   NULL, /* comment */
   NULL, /* comment_dot */
   NULL, /* comment_filepos */
+  format_comment_special /* comment */
 };
 
 
@@ -743,5 +946,3 @@ add_mo_suffix (fname)
   stpcpy (stpcpy (result, fname), ".mo");
   return result;
 }
-
-
