@@ -1,5 +1,5 @@
 /* GNU gettext - internationalization aids
-   Copyright (C) 1995 Free Software Foundation, Inc.
+   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
 
    This file was written by Peter Miller <pmiller@agso.gov.au>
 
@@ -36,6 +36,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 # include <locale.h>
 #endif
 
+#include "dir-list.h"
 #include "error.h"
 #include "message.h"
 #include <system.h>
@@ -66,6 +67,11 @@ struct merge_class_ty
   string_list_ty *comment;
   string_list_ty *comment_dot;
 
+  /* Flags transported in special comments.  */
+  int is_fuzzy;
+  enum is_c_format is_c_format;
+  enum is_c_format do_wrap;
+
   /* Accumulate filepos comments for the next message directive.  */
   size_t filepos_count;
   lex_pos_ty *filepos;
@@ -75,44 +81,52 @@ struct merge_class_ty
 /* String containing name the program is called with.  */
 const char *program_name;
 
-/* Force output to be written even if errors are found.  */
-static int force_output;
+/* If non-zero do not print unneeded messages.  */
+static int quiet;
+
+/* Verbosity level.  */
+static int verbosity_level;
 
 /* Long options.  */
 static const struct option long_options[] =
 {
+  { "directory", required_argument, NULL, 'D' },
   { "escape", no_argument, NULL, 'E' },
-  { "force", no_argument, NULL, 'f' },
   { "help", no_argument, NULL, 'h' },
   { "indent", no_argument, NULL, 'i' },
   { "no-escape", no_argument, NULL, 'e' },
   { "output-file", required_argument, NULL, 'o' },
+  { "quiet", no_argument, NULL, 'q' },
   { "sort-by-file", no_argument, NULL, 'F' },
   { "sort-output", no_argument, NULL, 's' },
+  { "silent", no_argument, NULL, 'q' },
   { "strict", no_argument, NULL, 'S' },
+  { "verbose", no_argument, NULL, 'v' },
   { "version", no_argument, NULL, 'V' },
+  { "width", required_argument, NULL, 'w', },
   { NULL, 0, NULL, 0 }
 };
 
 
 /* Prototypes for local functions.  */
-static void usage __P ((int __status));
-static void error_print __P ((void));
-static void merge_constructor __P ((po_ty *__that));
-static void merge_destructor __P ((po_ty *__that));
-static void merge_directive_domain __P ((po_ty *__that, char *__name));
-static void merge_directive_message __P ((po_ty *__that, char *__msgid,
-					  lex_pos_ty *__msgid_pos,
-					  char *__msgstr,
-					  lex_pos_ty *__msgstr_pos));
-static void merge_parse_brief __P ((po_ty *__that));
-static void merge_parse_debrief __P ((po_ty *__that));
-static void merge_comment __P ((po_ty *__that, const char *__s));
-static void merge_comment_dot __P ((po_ty *__that, const char *__s));
-static void merge_comment_filepos __P ((po_ty *__that, const char *__name,
-					int __line));
-static message_list_ty *grammar __P ((char *__filename));
-static message_list_ty *merge __P ((char *__fn1, char *__fn2));
+static void usage PARAMS ((int __status));
+static void error_print PARAMS ((void));
+static void merge_constructor PARAMS ((po_ty *__that));
+static void merge_destructor PARAMS ((po_ty *__that));
+static void merge_directive_domain PARAMS ((po_ty *__that, char *__name));
+static void merge_directive_message PARAMS ((po_ty *__that, char *__msgid,
+					     lex_pos_ty *__msgid_pos,
+					     char *__msgstr,
+					     lex_pos_ty *__msgstr_pos));
+static void merge_parse_brief PARAMS ((po_ty *__that));
+static void merge_parse_debrief PARAMS ((po_ty *__that));
+static void merge_comment PARAMS ((po_ty *__that, const char *__s));
+static void merge_comment_dot PARAMS ((po_ty *__that, const char *__s));
+static void merge_comment_special PARAMS ((po_ty *__that, const char *__s));
+static void merge_comment_filepos PARAMS ((po_ty *__that, const char *__name,
+					   int __line));
+static message_list_ty *grammar PARAMS ((const char *__filename));
+static message_list_ty *merge PARAMS ((const char *__fn1, const char *__fn2));
 
 
 int
@@ -120,7 +134,7 @@ main (argc, argv)
      int argc;
      char **argv;
 {
-  int optchar;
+  int opt;
   int do_help;
   int do_version;
   char *output_file;
@@ -130,7 +144,10 @@ main (argc, argv)
 
   /* Set program name for messages.  */
   program_name = argv[0];
+  verbosity_level = 0;
+  quiet = 0;
   error_print_progname = error_print;
+  gram_max_allowed_errors = INT_MAX;
 
 #ifdef HAVE_SETLOCALE
   /* Set locale via LC_ALL.  */
@@ -145,15 +162,19 @@ main (argc, argv)
   do_help = 0;
   do_version = 0;
   output_file = NULL;
-  force_output = 0;
 
-  while ((optchar = getopt_long (argc, argv, "eEfFhio:sV", long_options, NULL))
+  while ((opt
+	  = getopt_long (argc, argv, "D:eEFhio:qsvVw:", long_options, NULL))
 	 != EOF)
-    switch (optchar)
+    switch (opt)
       {
       case '\0':		/* Long option.  */
 	break;
 
+      case 'D':
+	dir_list_append (optarg);
+	break;
+
       case 'e':
 	message_print_style_escape (0);
 	break;
@@ -162,12 +183,6 @@ main (argc, argv)
 	message_print_style_escape (1);
 	break;
 
-      case 'f':
-	/* We don't want the program to abort because there are errors.  */
-	gram_max_allowed_errors = INT_MAX;
-	force_output = 1;
-	break;
-
       case 'F':
         sort_by_filepos = 1;
         break;
@@ -180,14 +195,14 @@ main (argc, argv)
 	message_print_style_indent ();
 	break;
 
-      case 'V':
-	do_version = 1;
-	break;
-
       case 'o':
 	output_file = optarg;
 	break;
 
+      case 'q':
+	quiet = 1;
+	break;
+
       case 's':
         sort_by_msgid = 1;
         break;
@@ -196,6 +211,24 @@ main (argc, argv)
 	message_print_style_uniforum ();
 	break;
 
+      case 'v':
+	++verbosity_level;
+	break;
+
+      case 'V':
+	do_version = 1;
+	break;
+
+      case 'w':
+	{
+	  int value;
+	  char *endp;
+	  value = strtol (optarg, &endp, 10);
+	  if (endp != optarg)
+	    message_page_width_set (value);
+	}
+	break;
+
       default:
 	usage (EXIT_FAILURE);
 	break;
@@ -234,7 +267,7 @@ main (argc, argv)
     message_list_sort_by_msgid (result);
 
   /* Write the merged message list out.  */
-  message_list_print (result, output_file);
+  message_list_print (result, output_file, 0, 0);
 
   exit (EXIT_SUCCESS);
 }
@@ -250,17 +283,22 @@ usage (status)
 	     program_name);
   else
     {
+      /* xgettext: no-wrap */
       printf (_("\
 Usage: %s [OPTION] def.po ref.po\n\
 Mandatory arguments to long options are mandatory for short options too.\n\
-  -e, --no-escape          do not use C escapes in output (default)\n\
-  -E, --escape             use C escapes in output, no extended chars\n\
-  -h, --help               display this help and exit\n\
-  -i, --indent             indented output style\n\
-  -o, --output-file=FILE   result will be written to FILE\n\
-  -S, --strict             strict Uniforum output style\n\
-  -V, --version            output version information and exit\n"),
+  -D, --directory=DIRECTORY   add DIRECTORY to list for input files search\n\
+  -e, --no-escape             do not use C escapes in output (default)\n\
+  -E, --escape                use C escapes in output, no extended chars\n\
+  -h, --help                  display this help and exit\n\
+  -i, --indent                indented output style\n\
+  -o, --output-file=FILE      result will be written to FILE\n\
+      --strict                strict Uniforum output style\n\
+  -v, --verbose               increase verbosity level\n\
+  -V, --version               output version information and exit\n\
+  -w, --width=NUMBER          set output page width\n"),
 		program_name);
+      /* xgettext: no-wrap */
       fputs (_("\n\
 Merges two Uniforum style .po files together.  The def.po file is an\n\
 existing PO file with the old translations which will be taken over to\n\
@@ -304,6 +342,9 @@ merge_constructor (that)
   this->comment_dot = NULL;
   this->filepos_count = 0;
   this->filepos = NULL;
+  this->is_fuzzy = 0;
+  this->is_c_format = undecided;
+  this->do_wrap = undecided;
 }
 
 
@@ -410,17 +451,24 @@ merge_directive_message (that, msgid, msgid_pos, msgstr, msgstr_pos)
       message_comment_filepos (mp, pp->file_name, pp->line_number);
       free (pp->file_name);
     }
+  mp->is_fuzzy = this->is_fuzzy;
+  mp->is_c_format = this->is_c_format;
+  mp->do_wrap = this->do_wrap;
+
   if (this->filepos != NULL)
     free (this->filepos);
   this->filepos_count = 0;
   this->filepos = NULL;
+  this->is_fuzzy = 0;
+  this->is_c_format = undecided;
+  this->do_wrap = undecided;
 
   /* See if this domain has been seen for this message ID.  */
   mvp = message_variant_search (mp, this->domain);
   if (mvp)
     {
-      gram_error_with_loc (msgid_pos, _("duplicate message definition"));
-      gram_error_with_loc (&mvp->pos, _("\
+      gram_error_at_line (msgid_pos, _("duplicate message definition"));
+      gram_error_at_line (&mvp->pos, _("\
 this is the location of the first definition"));
       free (msgstr);
     }
@@ -468,7 +516,7 @@ merge_parse_debrief (that)
 		break;
 	    }
 	  if (m >= mp->variant_count)
-	    gram_error_with_loc (&mp->variant[0].pos, _("\
+	    gram_error_at_line (&mp->variant[0].pos, _("\
 this message has no definition in the \"%s\" domain"), domain_name);
 	}
     }
@@ -501,6 +549,21 @@ merge_comment_dot (that, s)
 }
 
 
+static void
+merge_comment_special (that, s)
+     po_ty *that;
+     const char *s;
+{
+  merge_class_ty *this = (merge_class_ty *) that;
+
+  if (strstr (s, "fuzzy") != NULL)
+    this->is_fuzzy = 1;
+
+  this->is_c_format = parse_c_format_description_string (s);
+  this->do_wrap = parse_c_width_description_string (s);
+}
+
+
 static void
 merge_comment_filepos (that, name, line)
      po_ty *that;
@@ -537,17 +600,19 @@ static po_method_ty merge_methods =
   merge_comment,
   merge_comment_dot,
   merge_comment_filepos,
+  merge_comment_special
 };
 
 
 static message_list_ty *
 grammar (filename)
-     char *filename;
+     const char *filename;
 {
   po_ty *pop;
   message_list_ty *mlp;
 
   pop = po_alloc (&merge_methods);
+  po_lex_pass_obsolete_entries (1);
   po_scan (pop, filename);
   mlp = ((merge_class_ty *) pop)->mlp;
   po_free (pop);
@@ -555,18 +620,22 @@ grammar (filename)
 }
 
 
+#define DOT_FREQUENCE 10
+
 static message_list_ty *
 merge (fn1, fn2)
-     char *fn1;			/* definitions */
-     char *fn2;			/* references */
+     const char *fn1;			/* definitions */
+     const char *fn2;			/* references */
 {
   message_list_ty *def;
   message_list_ty *ref;
-  int nerrors;
   message_ty *defmsg;
   size_t j, k;
+  size_t merged, fuzzied, missing, obsolete;
   message_list_ty *result;
 
+  merged = fuzzied = missing = obsolete = 0;
+
   /* This is the definitions file, created by a human.  */
   def = grammar (fn1);
 
@@ -577,11 +646,15 @@ merge (fn1, fn2)
   result = message_list_alloc ();
 
   /* Every reference must be matched with its definition. */
-  nerrors = 0;
   for (j = 0; j < ref->nitems; ++j)
     {
       message_ty *refmsg;
 
+      /* Because merging can take a while we print something to signal
+	 we are not dead.  */
+      if (!quiet && verbosity_level <= 1 && j % DOT_FREQUENCE == 0)
+	fputc ('.', stderr);
+
       refmsg = ref->item[j];
 
       /* See if it is in the other file.  */
@@ -593,27 +666,31 @@ merge (fn1, fn2)
 	     the definition, take the msgstr from the definition.  Add
 	     this merged entry to the output message list.  */
 	  message_ty *mp = message_merge (defmsg, refmsg);
+
 	  message_list_append (result, mp);
 
 	  /* Remember that this message has been used, when we scan
 	     later to see if anything was omitted.  */
 	  defmsg->used = 1;
+	  ++merged;
 	  continue;
 	}
 
       /* If the message was not defined at all, try to find a very
 	 similar message, it could be a typo, or the suggestion may
 	 help.  */
-      ++nerrors;
       defmsg = message_list_search_fuzzy (def, refmsg->msgid);
       if (defmsg)
 	{
 	  message_ty *mp;
 
-	  gram_error_with_loc (&refmsg->variant[0].pos, _("\
+	  if (verbosity_level > 1)
+	    {
+	      gram_error_at_line (&refmsg->variant[0].pos, _("\
 this message is used but not defined..."));
-	  gram_error_with_loc (&defmsg->variant[0].pos, _("\
+	      gram_error_at_line (&defmsg->variant[0].pos, _("\
 ...but this definition is similar"));
+	    }
 
 	  /* Merge the reference with the definition: take the #. and
 	     #: comments from the reference, take the # comments from
@@ -621,23 +698,30 @@ this message is used but not defined..."));
 	     this merged entry to the output message list.  */
 	  mp = message_merge (defmsg, refmsg);
 
-	  /* Add a comment in case we write the file out.  */
-	  message_comment_special_append (mp, "INEXACT");
+	  mp->is_fuzzy = 1;
 
 	  message_list_append (result, mp);
 
 	  /* Remember that this message has been used, when we scan
 	     later to see if anything was omitted.  */
 	  defmsg->used = 1;
+	  ++fuzzied;
+	  if (!quiet && verbosity_level <= 1)
+	    /* Always print a dot if we handled a fuzzy match.  */
+	    fputc ('.', stderr);
 	}
       else
 	{
 	  message_ty *mp;
 
-	  gram_error_with_loc (&refmsg->variant[0].pos, _("\
+	  if (verbosity_level > 1)
+	      gram_error_at_line (&refmsg->variant[0].pos, _("\
 this message is used but not defined in %s"), fn1);
+
 	  mp = message_copy (refmsg);
+
 	  message_list_append (result, mp);
+	  ++missing;
 	}
     }
 
@@ -649,15 +733,24 @@ this message is used but not defined in %s"), fn1);
       defmsg = def->item[k];
       if (defmsg->used)
 	continue;
-      gram_error_with_loc (&defmsg->variant[0].pos,
-			   _("warning: this message is not used"));
+
+      /* Remember the old translation although it is not used anymore.
+	 But we mark it as obsolete.  */
+      defmsg->obsolete = 1;
+
+      message_list_append (result, defmsg);
+      ++obsolete;
     }
 
-  /* Exit with status 1 on any error unless we are forced to write the
-     result.  */
-  if (nerrors > 0)
-    error ((force_output ? 0 : EXIT_FAILURE), 0,
-	   _("found %d errors"), nerrors);
+  /* Report some statistics.  */
+  if (verbosity_level > 0)
+    fprintf (stderr, _("%s\
+Read %d old + %d reference, \
+merged %d, fuzzied %d, missing %d, obsolete %d.\n"),
+	     !quiet && verbosity_level <= 1 ? "\n" : "",
+	     def->nitems, ref->nitems, merged, fuzzied, missing, obsolete);
+  else if (!quiet)
+    fputs (_(" done.\n"), stderr);
 
   return result;
 }
